from .review import ReviewStrategy, SM2ReviewStrategy, ReviewStrategyFactory
from .storage import EntryStorageBase
from .vocabulary import Vocabulary

__all__ = [
    "ReviewStrategy",
    "SM2ReviewStrategy",
    "ReviewStrategyFactory",
    "EntryStorageBase",
    "Vocabulary",
]



from abc import ABC, abstractmethod

from langcodes import Language

from verbia_core.entry import Entry


class DictionaryBase(ABC):
    def __init__(self, source: str):
        self._source = source

    @abstractmethod
    def lookup(
        self, word: str, word_language: Language, native_language: Language
    ) -> Entry:
        pass

    @abstractmethod
    async def async_lookup(
        self, word: str, word_language: Language, native_language: Language
    ) -> Entry:
        pass



import uuid
from dataclasses import dataclass, field

from langcodes import Language

from verbia_core.utils import time_provider


@dataclass
class Entry:
    word: str
    native_language: Language
    native_language_definition: str
    source: str
    is_new: bool
    word_language: Language

    example_sentences: list[str] = field(default_factory=list)
    notes: list[str] = field(default_factory=list)

    created_at: int = time_provider.time_mills_from_now()
    next_review_at: int = time_provider.time_mills_from_now(interval_days=1)
    review_interval_days: int = 1
    interval: int = 1
    repetitions: int = 0
    quality: int = 0
    ease_factor: float = 2.5
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    vocabulary_id: str = ""



from __future__ import annotations

from os import environ

from verbia_core.error import VerbiaError


def get_client(api_key: str):
    import google.generativeai as genai

    if not api_key:
        api_key = environ.get("GEMINI_API_KEY")
    if not api_key:
        raise VerbiaError("GEMINI_API_KEY required.")
    model_name = environ.get("GEMINI_MODEL_NAME", "gemini-1.5-flash")
    genai.configure(api_key=api_key)
    return genai.GenerativeModel(model_name=model_name)



from unittest.mock import MagicMock, patch, PropertyMock, AsyncMock

import pytest
from langcodes import Language

from verbia_core.entry import Entry
from verbia_core.error import VerbiaError
from verbia_core.vocabulary import Vocabulary


class FakeVocabulary(Vocabulary):
    @property
    def _entry_storage(self):
        return MagicMock()

    @property
    def _dictionary(self):
        return MagicMock()

    @property
    def _review_strategy(self):
        return MagicMock()

    def self_delete(self):
        pass


@pytest.fixture
def mock_vocabulary():
    vocab = FakeVocabulary(
        id="vocab1",
        name="Test Vocabulary",
        word_language=Language.get("en"),
        native_language=Language.get("fr"),
    )
    return vocab


@pytest.fixture
def mock_entry():
    return Entry(
        word="test",
        native_language=Language.get("fr"),
        word_language=Language.get("en"),
        source="Gemini",
        is_new=False,
        native_language_definition="test definition",
        vocabulary_id="vocab1",
    )


@patch.object(FakeVocabulary, "_entry_storage", new_callable=PropertyMock)
def test_add_word_existing_entry(_entry_storage, mock_vocabulary, mock_entry):
    mock_entry_storage = MagicMock()
    mock_entry_storage.get.return_value = mock_entry
    _entry_storage.return_value = mock_entry_storage

    result = mock_vocabulary.add_word("test")

    assert result == mock_entry

    mock_entry_storage.get.assert_called_once_with(mock_entry.word, mock_vocabulary.id)
    mock_entry_storage.add_or_update.assert_not_called()


@patch.object(FakeVocabulary, "_entry_storage", new_callable=PropertyMock)
@patch.object(FakeVocabulary, "_dictionary", new_callable=PropertyMock)
def test_add_word_new_entry(_dictionary, _entry_storage, mock_vocabulary, mock_entry):
    mock_entry_storage = MagicMock()
    mock_dictionary = MagicMock()
    _entry_storage.return_value = mock_entry_storage
    _dictionary.return_value = mock_dictionary
    mock_entry_storage.get.return_value = None
    mock_dictionary.lookup.return_value = mock_entry

    result = mock_vocabulary.add_word("test")

    assert result == mock_entry
    mock_entry_storage.get.assert_called_once_with(mock_entry.word, mock_vocabulary.id)
    mock_entry_storage.add_or_update.assert_called_once_with(mock_entry)
    mock_dictionary.lookup.assert_called_once_with(
        mock_entry.word, mock_entry.word_language, mock_entry.native_language
    )


@patch.object(FakeVocabulary, "_entry_storage", new_callable=PropertyMock)
@patch.object(FakeVocabulary, "_dictionary", new_callable=PropertyMock)
def test_add_word_raises_error(_dictionary, _entry_storage, mock_vocabulary):
    mock_entry_storage = MagicMock()
    mock_dictionary = MagicMock()
    _entry_storage.return_value = mock_entry_storage
    _dictionary.return_value = mock_dictionary
    mock_entry_storage.get.return_value = None
    mock_dictionary.lookup.return_value = None

    with pytest.raises(
        VerbiaError, match="Word 'unknown' does not exist in the dictionary."
    ):
        mock_vocabulary.add_word("unknown")


@pytest.mark.asyncio
@patch.object(FakeVocabulary, "_entry_storage", new_callable=PropertyMock)
@patch.object(FakeVocabulary, "_dictionary", new_callable=PropertyMock)
async def test_async_add_word_new_entry(
    _dictionary, _entry_storage, mock_vocabulary, mock_entry
):
    mock_entry_storage = AsyncMock()
    mock_dictionary = AsyncMock()
    _entry_storage.return_value = mock_entry_storage
    _dictionary.return_value = mock_dictionary
    mock_entry_storage.async_get.return_value = None
    mock_dictionary.async_lookup.return_value = mock_entry

    result = await mock_vocabulary.async_add_word("test")

    assert result == mock_entry

    mock_entry_storage.async_get.assert_called_once_with(
        mock_entry.word, mock_vocabulary.id
    )
    mock_entry_storage.async_add_or_update.assert_awaited_once_with(mock_entry)
    mock_dictionary.async_lookup.assert_awaited_once_with(
        mock_entry.word, mock_entry.word_language, mock_entry.native_language
    )


@patch.object(FakeVocabulary, "_entry_storage", new_callable=PropertyMock)
@patch.object(FakeVocabulary, "_review_strategy", new_callable=PropertyMock)
def test_update_review(_review_strategy, _entry_storage, mock_vocabulary, mock_entry):
    mock_entry_storage = MagicMock()
    mock_review_strategy = MagicMock()
    _entry_storage.return_value = mock_entry_storage
    _review_strategy.return_value = mock_review_strategy

    updated_entry = Entry(
        word=mock_entry.word,
        native_language=mock_entry.native_language,
        word_language=mock_entry.word_language,
        source=mock_entry.source,
        is_new=mock_entry.is_new,
        native_language_definition=mock_entry.native_language_definition,
        vocabulary_id=mock_entry.vocabulary_id,
        next_review_at=12345,
        repetitions=mock_entry.repetitions + 1,
    )

    mock_review_strategy.update_review.return_value = updated_entry

    mock_vocabulary.update_review(mock_entry, 5)

    mock_review_strategy.update_review.assert_called_once_with(mock_entry, 5)
    mock_entry_storage.add_or_update.assert_called_once_with(updated_entry)



import json
import os
from string import Template

from langcodes import Language
from loguru import logger

from verbia_core.dictionary.base import DictionaryBase
from verbia_core.dictionary.gemini.client import get_client
from verbia_core.entry import (
    EnglishEntry,
    Forms,
    Entry,
    JapaneseEntry,
    JapaneseReading,
    Conjugation,
)
from verbia_core.error import WordInvalidError

# Suppress logging warnings
os.environ["GRPC_VERBOSITY"] = "ERROR"
os.environ["GLOG_minloglevel"] = "2"

english_prompt = Template("""You are a language assistant for looking up English words. When I provide a word, please give me the following information in $native_language:

1. The definition or meaning of the word in $native_language, along with its part of speech (e.g., n. for noun, v. for verb, adj. for adjective).
2. Example sentences in English demonstrating how the word is used in context.
3. If the word has multiple forms (e.g., plural, past tense), provide those forms in $native_language.
4. The lemma (base form) of the word, if applicable.
5. The pronunciation of the word in IPA format, if available.

Make sure the response follows this JSON structure:
{
  "word": "{word}",
  "definition": "{definition_in_native_language}",
  "example_sentences": [
    "{sentence_1}",
    "{sentence_2}",
    "{sentence_3}"
  ],
  "forms": {
        "past_tense": "{past_tense_form}",
        "present_participle": "{present_participle_form}",
        "past_participle": "{past_participle_form}",
        "third_person_singular": "{third_person_singular_form}",
        "singular": "{singular_form}",
        "plural": "{plural_form}",
        "comparative": "{comparative_form}",
        "superlative": "{superlative_form}"
    },
  "lemma": "{lemma}",
  "pronunciation": "{IPA_pronunciation}"
}

Here is the word I want to look up: $word
If the word is not a valid English word, please respond null as the value for field "definition".
""")

japanese_prompt = Template("""
You are a language assistant for looking up Japanese words. When I provide a word, please give me the following information in $native_language:

1. The definition or meaning of the word in $native_language.
2. The readings (hiragana/katakana/kanji) of the word, including any kanji readings (kun'yomi and on'yomi).
3. Example sentences in Japanese demonstrating how the word is used in context.
4. If the word is a verb, provide its conjugation in different tenses and forms (e.g., present, past, negative form, te-form, potential form, polite form).

Make sure the response follows this JSON structure:
{
    "word": "[word]",
    "definition": "[definition_in_native_language]",
    "reading": {
        "hiragana": "[hiragana_reading]",
        "katakana": "[katakana_reading]",
        "kanji": {
            "kunyomi": "[kunyomi_reading]",
            "onyomi": "[onyomi_reading]"
    },
    "example_sentences": [
        "[example_sentence_1]",
        "[example_sentence_2]"
    ],
    "conjugation": {
        "present": "[present_conjugation]",
        "past": "[past_conjugation]",
        "negative": "[negative_conjugation]",
        "te-form": "[te_form]",
        "potential": "[potential_form]",
        "polite": "[polite_form]"
    }
}

Conjugation should be provided only if the word is a verb and contains the original word. 
For example, if the word is "勉強", the conjugation should be provided for "勉強する" and not "する".

Here is the word I want to look up: $word
If the word is not a valid Japanese word, please respond null as the value for field "definition".
""")

common_language_prompt = Template("""
You are a language assistant for looking up words in $word_language. When I provide a word, please give me the following information in $native_language:

1. The definition or meaning of the word in $native_language.
2. Example sentences in $word_language demonstrating how the word is used in context.

Make sure the response follows this JSON structure:
{
    "word": "[word]",
    "definition": "[definition_in_native_language]",
    "example_sentences": [
        "[example_sentence_1]",
        "[example_sentence_2]"
    ]
}

Here is the word I want to look up: $word
If the word is not a valid word in $word_language, please respond null as the value for field "definition".
""")


def _prepare_prompt(
    word: str, word_language: Language, native_language: Language
) -> str:
    match word_language.language:
        case "en":
            return english_prompt.substitute(
                word=word, native_language=native_language.display_name()
            )
        case "ja":
            return japanese_prompt.substitute(
                word=word, native_language=native_language.display_name()
            )
        case _:
            return common_language_prompt.substitute(
                word=word,
                word_language=word_language.display_name(),
                native_language=native_language.display_name(),
            )


def _is_none(value: str) -> bool:
    return (
        value is None
        or value == "null"
        or value == ""
        or value == "None"
        or value == "none"
        or value == "undefined"
        or value == "nil"
        or value == "NaN"
        or value == "nan"
        or value == "NULL"
        or value == "NoneType"
        or value == "noneType"
        or value == "undefinedType"
        or value == "nilType"
        or value == "NaNType"
        or value == "nanType"
    )


class GeminiDictionary(DictionaryBase):
    def __init__(self, api_key: str | None = None):
        self.__client = None
        self._api_key = api_key
        super().__init__("Gemini")

    @property
    def _client(self):
        if not self.__client:
            self.__client = get_client(self._api_key)
        return self.__client

    def _generate(self, prompt: str) -> dict:
        import google.generativeai as genai

        result = self._client.generate_content(
            contents=prompt,
            generation_config=genai.GenerationConfig(
                response_mime_type="application/json"
            ),
        )
        return json.loads(result.text)

    async def _async_generate(self, prompt: str) -> dict:
        import google.generativeai as genai

        result = await self._client.generate_content_async(
            contents=prompt,
            generation_config=genai.GenerationConfig(
                response_mime_type="application/json"
            ),
        )
        return json.loads(result.text)

    def _extract_to_entry(
        self,
        response: dict,
        word: str,
        word_language: Language,
        native_language: Language,
    ) -> Entry:
        match word_language.language:
            case "en":
                return self._extract_to_english_entry(response, word, native_language)
            case "ja":
                return self._extract_to_japanese_entry(response, word, native_language)
            case _:
                return self._extract_to_common_entry(
                    response, word, word_language, native_language
                )

    def _extract_to_english_entry(
        self, response: dict, word: str, native_language: Language
    ) -> EnglishEntry:
        if not isinstance(response, dict):
            logger.error("Response is not a valid dictionary")
            raise ValueError("Invalid response format: expected a dictionary")

        definition = response.get("definition")
        if _is_none(definition):
            logger.warning(f"Definition not found for word: {word}")
            raise WordInvalidError(word, native_language.display_name())

        _forms = response.get("forms", {})
        if not isinstance(_forms, dict):
            logger.warning(
                f"Forms for word '{word}' should be a dictionary. Using empty dictionary."
            )
            forms = None
        else:
            try:
                past_tense = _forms.get("past_tense")
                present_participle = _forms.get("present_participle")
                past_participle = _forms.get("past_participle")
                third_person_singular = _forms.get("third_person_singular")
                singular = _forms.get("singular")
                plural = _forms.get("plural")
                comparative = _forms.get("comparative")
                superlative = _forms.get("superlative")
                forms = Forms(
                    past_tense=past_tense,
                    present_participle=present_participle,
                    past_participle=past_participle,
                    third_person_singular=third_person_singular,
                    singular=singular,
                    plural=plural,
                    comparative=comparative,
                    superlative=superlative,
                )
            except Exception as e:
                logger.error(f"Error processing forms for word '{word}': {str(e)}")
                forms = None

        lemma = response.get("lemma")
        pronunciation = response.get("pronunciation")

        return EnglishEntry(
            word=word,
            native_language=native_language,
            native_language_definition=definition,
            forms=forms,
            lemma=lemma,
            pronunciation=pronunciation,
            source=self._source,
            is_new=True,
        )

    def _extract_to_japanese_entry(
        self, response: dict, word: str, native_language: Language
    ) -> JapaneseEntry:
        if not isinstance(response, dict):
            logger.error("Response is not a valid dictionary")
            raise ValueError("Invalid response format: expected a dictionary")

        definition = response.get("definition")
        if _is_none(definition):
            logger.warning(f"Definition not found for word: {word}")
            raise WordInvalidError(word, native_language.display_name())

        example_sentences = response.get("example_sentences", [])
        if not isinstance(example_sentences, list):
            logger.warning(
                f"Example sentences for word '{word}' should be a list. Using empty list."
            )
            example_sentences = []

        _reading = response.get("reading")
        if not isinstance(_reading, dict):
            logger.warning(
                f"Reading for word '{word}' should be a dictionary. Using empty dictionary."
            )
            reading = None
        else:
            try:
                hiragana = _reading.get("hiragana")
                katakana = _reading.get("katakana")
                kanji = _reading.get("kanji")
                kunyomi = kanji.get("kunyomi") if kanji else None
                onyomi = kanji.get("onyomi") if kanji else None
                reading = JapaneseReading(
                    hiragana=hiragana, katakana=katakana, kunyomi=kunyomi, onyomi=onyomi
                )
            except Exception as e:
                logger.error(f"Error processing reading for word '{word}': {str(e)}")
                reading = None

        _conjugation = response.get("conjugation")
        if not isinstance(_conjugation, dict):
            logger.warning(
                f"Conjugation for word '{word}' should be a dictionary. Using empty dictionary."
            )
            conjugation = None
        else:
            try:
                present = _conjugation.get("present")
                past = _conjugation.get("past")
                negative = _conjugation.get("negative")
                te_form = _conjugation.get("te-form")
                potential = _conjugation.get("potential")
                polite = _conjugation.get("polite")
                conjugation = Conjugation(
                    present=present,
                    past=past,
                    negative=negative,
                    te_form=te_form,
                    potential=potential,
                    polite=polite,
                )
            except Exception as e:
                logger.error(
                    f"Error processing conjugation for word '{word}': {str(e)}"
                )
                conjugation = None

        return JapaneseEntry(
            word=word,
            native_language=native_language,
            native_language_definition=definition,
            example_sentences=example_sentences,
            reading=reading,
            conjugation=conjugation,
            source=self._source,
            is_new=True,
        )

    def _extract_to_common_entry(
        self,
        response: dict,
        word: str,
        word_language: Language,
        native_language: Language,
    ) -> Entry:
        if not isinstance(response, dict):
            logger.error("Response is not a valid dictionary")
            raise ValueError("Invalid response format: expected a dictionary")

        definition = response.get("definition")
        if _is_none(definition):
            logger.warning(f"Definition not found for word: {word}")
            raise WordInvalidError(word, native_language.display_name())

        example_sentences = response.get("example_sentences", [])
        if not isinstance(example_sentences, list):
            logger.warning(
                f"Example sentences for word '{word}' should be a list. Using empty list."
            )
            example_sentences = []

        return Entry(
            word=word,
            word_language=word_language,
            native_language=native_language,
            native_language_definition=definition,
            example_sentences=example_sentences,
            source=self._source,
            is_new=True,
        )

    def lookup(
        self, word: str, word_language: Language, native_language: Language
    ) -> Entry:
        logger.debug(
            f"Looking up word '{word}' in {native_language.display_name()} by {self._source}"
        )
        prompt = _prepare_prompt(word, word_language, native_language)
        response = self._generate(prompt)
        return self._extract_to_entry(response, word, word_language, native_language)

    async def async_lookup(
        self, word: str, word_language: Language, native_language: Language
    ) -> Entry:
        logger.debug(
            f"Looking up word '{word}' in {native_language.display_name()} by {self._source}"
        )
        prompt = _prepare_prompt(word, word_language, native_language)
        response = await self._async_generate(prompt)
        return self._extract_to_entry(response, word, word_language, native_language)



from .base import Entry
from .english_entry import EnglishEntry, Forms
from .japanese_entry import JapaneseEntry, JapaneseReading, Conjugation

__all__ = [
    "Entry",
    "EnglishEntry",
    "Forms",
    "JapaneseEntry",
    "JapaneseReading",
    "Conjugation",
]



from abc import ABC, abstractmethod

from verbia_core.entry import Entry


class EntryStorageBase(ABC):
    @abstractmethod
    def get(self, word: str, vocabulary_id: str) -> Entry | None:
        pass

    @abstractmethod
    def add_or_update(self, entry: Entry):
        pass

    @abstractmethod
    def delete(self, entry: Entry):
        pass

    @abstractmethod
    def list_all(self, vocabulary_id: str) -> list[Entry]:
        pass

    @abstractmethod
    def list_all_due(self, vocabulary_id) -> list[Entry]:
        pass

    @abstractmethod
    async def async_add_or_update(self, entry: Entry):
        pass

    @abstractmethod
    async def async_get(self, word: str, vocabulary_id: str) -> Entry | None:
        pass

    @abstractmethod
    def delete_by_vocabulary_id(self, vocabulary_id: str):
        pass



from verbia_core.dictionary.base import DictionaryBase
from verbia_core.dictionary.gemini.dictionary import GeminiDictionary


class DictionaryFactory:
    _dictionaries = {
        "Gemini": GeminiDictionary,
    }

    @classmethod
    def create(cls, dictionary_name: str, **kwargs) -> DictionaryBase:
        dictionary_cls = cls._dictionaries.get(dictionary_name)
        if not dictionary_cls:
            raise ValueError(f"Dictionary '{dictionary_name}' not found.")
        return dictionary_cls(**kwargs)



from abc import ABC, abstractmethod

from verbia_core.entry import Entry
from verbia_core.utils import time_provider


class ReviewStrategy(ABC):
    def __init__(self, name: str):
        self._name = name

    @abstractmethod
    def update_review(self, entry: Entry, quality: int) -> Entry:
        raise NotImplementedError


class SM2ReviewStrategy(ReviewStrategy):
    def __init__(self):
        super().__init__("SM2")

    def update_review(self, entry: Entry, quality: int) -> Entry:
        entry.ease_factor += 0.1 - (5 - quality) * (0.08 + (5 - quality) * 0.02)
        entry.ease_factor = max(entry.ease_factor, 1.3)  # 确保易度因子不小于 1.3

        if entry.repetitions == 0:
            entry.interval = 1
        elif entry.repetitions == 1:
            entry.interval = 6
        else:
            entry.interval = round(entry.interval * entry.ease_factor)

        entry.next_review_date = time_provider.time_mills_from_now(entry.interval)
        entry.repetitions += 1

        return entry


class ReviewStrategyFactory:
    _strategies = {
        "SM2": SM2ReviewStrategy,
    }

    @classmethod
    def create(cls, strategy_name: str) -> ReviewStrategy:
        strategy_cls = cls._strategies.get(strategy_name)
        if strategy_cls is None:
            raise ValueError(f"Unknown review strategy: {strategy_name}")
        return strategy_cls()



from dataclasses import dataclass

from langcodes import Language

from verbia_core.entry.base import Entry


@dataclass()
class Conjugation:
    present: str | None = None
    past: str | None = None
    negative: str | None = None
    te_form: str | None = None
    potential: str | None = None
    polite: str | None = None


@dataclass
class JapaneseReading:
    hiragana: str | None = None
    katakana: str | None = None
    kunyomi: str | None = None
    onyomi: str | None = None


@dataclass
class JapaneseEntry(Entry):
    word_language: Language = Language.get("ja")
    reading: JapaneseReading | None = None
    conjugation: Conjugation | None = None



class VerbiaError(Exception):
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return self.message


class WordInvalidError(VerbiaError):
    def __init__(self, word: str, language: str):
        super().__init__(f"Word '{word}' is not valid in language {language}")



from abc import abstractmethod, ABC

from langcodes import Language
from loguru import logger

from verbia_core.dictionary import DictionaryBase
from verbia_core.entry import EnglishEntry, Entry
from verbia_core.error import VerbiaError
from verbia_core.vocabulary.review import ReviewStrategy
from verbia_core.vocabulary.storage import EntryStorageBase


class Vocabulary(ABC):
    def __init__(
        self,
        id: str,
        name: str,
        word_language: Language,
        native_language: Language,
    ):
        self.id = id
        self.name = name
        self.word_language = word_language
        self.native_language = native_language

    def __str__(self):
        return f"Vocabulary(id={self.id}, name={self.name}, word_language={self.word_language}, native_language={self.native_language})"

    @property
    @abstractmethod
    def _entry_storage(self) -> EntryStorageBase:
        pass

    @property
    @abstractmethod
    def _dictionary(self) -> DictionaryBase:
        pass

    @property
    @abstractmethod
    def _review_strategy(self) -> ReviewStrategy:
        pass

    @abstractmethod
    def self_delete(self):
        pass

    def _lookup_word(self, word: str) -> Entry | None:
        logger.trace(f"Looking up word '{word}' in the dictionary")
        entry = self._dictionary.lookup(word, self.word_language, self.native_language)
        if not entry:
            raise VerbiaError(f"Word '{word}' does not exist in the dictionary.")
        entry.vocabulary_id = self.id

        return entry

    def add_word(self, word: str) -> Entry:
        logger.trace(f"Adding word '{word}' to vocabulary '{self.name}'")
        entry = self.get_entry(word)
        if not entry:
            entry = self._lookup_word(word)
            self.add_or_update_entry(entry)

        self._add_lemma_if_applicable(entry, word)
        return entry

    def _add_lemma_if_applicable(self, entry: Entry, word: str):
        if entry.is_new and isinstance(entry, EnglishEntry):
            en_entry: EnglishEntry = entry
            lemma = en_entry.lemma
            try:
                if lemma is not None and lemma != word:
                    self.add_word(lemma)
            except VerbiaError:
                pass

    def get_entry(self, word: str) -> Entry | None:
        logger.trace(f"Getting entry for word '{word}' from vocabulary '{self.name}'")
        entry = self._entry_storage.get(word, self.id)
        return entry

    def add_or_update_entry(self, entry: Entry):
        self._entry_storage.add_or_update(entry)

    def delete_entry(self, entry: Entry):
        self._entry_storage.delete(entry)

    def list_all_entries(self) -> list[Entry]:
        return self._entry_storage.list_all(self.id)

    def update_review(self, entry: Entry, quality: int):
        entry = self._review_strategy.update_review(entry, quality)
        self.add_or_update_entry(entry)

    def list_all_due_entries(self) -> list[Entry]:
        return self._entry_storage.list_all_due(self.id)

    async def async_get_entry(self, word: str) -> Entry | None:
        entry = await self._entry_storage.async_get(word, self.id)
        return entry

    async def async_lookup_word(self, word: str) -> Entry:
        entry = await self._dictionary.async_lookup(word, self.word_language, self.native_language)
        if not entry:
            raise VerbiaError(f"Word '{word}' does not exist in the dictionary.")
        entry.vocabulary_id = self.id

        return entry

    async def async_add_word(self, word: str):
        entry = await self.async_get_entry(word)
        if not entry:
            entry = await self.async_lookup_word(word)
            await self._entry_storage.async_add_or_update(entry)

        if entry.is_new and isinstance(entry, EnglishEntry):
            # Add lemma to vocabulary if it is different from the word, only for English words.
            en_entry: EnglishEntry = entry
            lemma = en_entry.lemma
            try:
                if lemma is not None and lemma != word:
                    await self.async_add_word(lemma)
            except VerbiaError:
                pass
        return entry



from dataclasses import dataclass

from langcodes import Language

from verbia_core.entry.base import Entry


@dataclass
class Forms:
    past_tense: str | None = None
    present_participle: str | None = None
    past_participle: str | None = None
    third_person_singular: str | None = None
    singular: str | None = None
    plural: str | None = None
    comparative: str | None = None
    superlative: str | None = None


@dataclass
class EnglishEntry(Entry):
    word_language: Language = Language.get("en")
    lemma: str | None = None
    forms: Forms | None = None
    pronunciation: str | None = None



from datetime import datetime, timedelta


def time_mills_from_now(interval_days: int = 0):
    """
    Returns the current time in milliseconds plus the interval in days.
    @param interval_days: The number of days to add to the current time.
    """

    now = datetime.now()
    future_time = now + timedelta(days=interval_days)
    return int(future_time.timestamp() * 1000)


def last_moment_of_day(interval_days: int = 0):
    """
    Returns the last moment of the day in milliseconds plus the interval in days.
    @param interval_days: The number of days to add to the current day.
    """

    now = datetime.now()
    future_time = now + timedelta(days=interval_days)
    future_time = future_time.replace(hour=23, minute=59, second=59, microsecond=999)
    return int(future_time.timestamp() * 1000)


def format_timestamp(timestamp: int) -> str:
    """Format timestamp to a readable date string."""
    return datetime.fromtimestamp(
        timestamp / 1000, tz=datetime.now().astimezone().tzinfo
    ).strftime("%Y-%m-%d %H:%M:%S")



from langcodes import Language

from verbia_core.dictionary.gemini.dictionary import GeminiDictionary
from verbia_core.entry import EnglishEntry, JapaneseEntry


def test_lookup_en():
    dictionary = GeminiDictionary()
    entry = dictionary.lookup(
        word="go",
        word_language=Language.get("en"),
        native_language=Language.get("zh"),
    )
    assert entry.word == "go"
    assert entry.word_language == Language.get("en")
    assert entry.native_language == Language.get("zh")
    assert entry.is_new is True
    assert entry.source == "Gemini"
    assert isinstance(entry, EnglishEntry)
    assert entry.lemma == "go"
    assert entry.forms.past_tense == "went"
    assert entry.forms.past_participle == "gone"
    assert entry.forms.present_participle == "going"
    assert entry.forms.third_person_singular == "goes"


def test_lookup_ja():
    dictionary = GeminiDictionary()
    entry = dictionary.lookup(
        word="勉強",
        word_language=Language.get("ja"),
        native_language=Language.get("en"),
    )
    assert entry.word == "勉強"
    assert entry.word_language == Language.get("ja")
    assert entry.native_language == Language.get("en")
    assert entry.is_new is True
    assert entry.source == "Gemini"
    assert isinstance(entry, JapaneseEntry)
    assert "べんきょう" in entry.reading.hiragana
    assert "勉強する" in entry.conjugation.present
    assert "勉強した" in entry.conjugation.past
    assert "勉強します" in entry.conjugation.polite
    assert "勉強しない" in entry.conjugation.negative
    assert "勉強して" in entry.conjugation.te_form
    assert "勉強できる" in entry.conjugation.potential
